from wagtail.contrib.redirects.models import Redirect
from rest_framework import serializers


class RedirectSerializer(serializers.ModelSerializer):
    class Meta:
        model = Redirect
        fields = ("old_path", "is_permanent", "site", "link")



import hashlib
from django.utils.timezone import datetime


def exclude_page_type(queryset, page_models):
    qs = queryset.none()
    for model in page_models:
        qs |= queryset.not_type(model)
    return qs


def hash_draft_code(code: str, page_id: int):
    """sha256 of date + code + page id"""
    date_code = datetime.today().strftime("%Y%m%d")
    combined_code = date_code + code + str(page_id)
    hashed_code = hashlib.sha256(combined_code.encode())
    return hashed_code.hexdigest()



from django.test import override_settings, RequestFactory
from wagtail.api.v2.router import WagtailAPIRouter
from wagtail.models import Page, Site
from wagtail.contrib.redirects.models import Redirect
from wagtail.test.utils import WagtailPageTests
from rest_framework.test import APIRequestFactory
from sandbox.models import FooPage
from .views import SPAExtendedPagesAPIEndpoint, RedirectViewSet, sitemap
from .utils import hash_draft_code


TEST_DRAFT_CODE = "abc"


class WagtailSPAIntegrationTests(WagtailPageTests):
    @override_settings(PREVIEW_DRAFT_CODE=TEST_DRAFT_CODE)
    def test_draft_api(self):
        home = Page.objects.last()
        old_title = home.title
        new_title = "edit it"
        home.title = new_title
        home.save_revision()

        request = APIRequestFactory().get("")
        request.site = Site.objects.first()
        request.wagtailapi_router = WagtailAPIRouter("wagtailapi")
        page_detail = SPAExtendedPagesAPIEndpoint.as_view({"get": "detail_view"})
        res = page_detail(request, pk=home.pk)
        self.assertContains(res, old_title)
        self.assertNotContains(res, new_title)

        params = {"draft": hash_draft_code(TEST_DRAFT_CODE, home.pk)}
        request = APIRequestFactory().get("", params)
        request.site = Site.objects.first()
        request.wagtailapi_router = WagtailAPIRouter("wagtailapi")
        res = page_detail(request, pk=home.pk)
        self.assertContains(res, new_title)

    @override_settings(PREVIEW_DRAFT_CODE=TEST_DRAFT_CODE)
    def test_draft_api_unpublished(self):
        home = Page.objects.last()
        new_title = "edit it"
        home.title = new_title
        home.save_revision()
        home.live = False
        home.save()

        params = {"draft": hash_draft_code(TEST_DRAFT_CODE, home.pk)}
        request = APIRequestFactory().get("", params)
        request.site = Site.objects.first()
        request.wagtailapi_router = WagtailAPIRouter("wagtailapi")
        page_detail = SPAExtendedPagesAPIEndpoint.as_view({"get": "detail_view"})
        res = page_detail(request, pk=home.pk)
        self.assertContains(res, new_title)

    @override_settings(PREVIEW_DRAFT_CODE=TEST_DRAFT_CODE)
    def test_draft_api_detail_by_path(self):
        home = Page.objects.last()
        foo = FooPage(title="foo")
        home.add_child(instance=foo)
        foo.live = False
        foo.save()

        url = "/api/v2/pages/detail_by_path/"
        params = {
            "html_path": foo.url,
            "draft": hash_draft_code(TEST_DRAFT_CODE, foo.pk),
        }
        res = self.client.get(url, params)
        self.assertContains(res, "foo")

    @override_settings(PREVIEW_DRAFT_CODE=TEST_DRAFT_CODE)
    def test_draft_api_detail_by_path_with_changed_slug(self):
        home = Page.objects.last()
        foo = FooPage(title="foo")
        home.add_child(instance=foo)
        foo.live = False
        foo.save()
        foo.slug = "foo-edit"
        foo.save_revision()

        url = "/api/v2/pages/detail_by_path/"
        params = {
            "html_path": "/foo-edit/",
            "draft": hash_draft_code(TEST_DRAFT_CODE, foo.pk),
        }
        res = self.client.get(url, params)
        self.assertContains(res, "foo")

    @override_settings(PREVIEW_DRAFT_CODE=TEST_DRAFT_CODE)
    def test_draft_api_detail_by_path_with_changed_slug_and_nested(self):
        """Ensure a deeply nested page works such as /sub/foo"""
        home = Page.objects.last()
        sub_page = FooPage(title="sub")
        home.add_child(instance=sub_page)
        foo = FooPage(title="foo")
        sub_page.add_child(instance=foo)
        foo.live = False
        foo.save()
        foo.slug = "foo-edit"
        foo.save_revision()

        url = "/api/v2/pages/detail_by_path/"
        params = {
            "html_path": "/foo-edit/",
            "draft": hash_draft_code(TEST_DRAFT_CODE, foo.pk),
        }
        res = self.client.get(url, params)
        self.assertContains(res, "foo")

    @override_settings(PREVIEW_DRAFT_CODE=TEST_DRAFT_CODE)
    def test_draft_api_detail_by_path_with_slug_404(self):
        """Pages detail_by_path 404s if no matching slug is found"""
        home = Page.objects.last()
        foo = FooPage(title="foo")
        home.add_child(instance=foo)
        foo.live = False
        foo.save()

        url = "/api/v2/pages/detail_by_path/"
        params = {
            "html_path": "/no/",
            "draft": hash_draft_code(TEST_DRAFT_CODE, foo.pk),
        }
        res = self.client.get(url, params)
        self.assertEquals(res.status_code, 404)

    @override_settings(PREVIEW_DRAFT_CODE=TEST_DRAFT_CODE)
    def test_draft_api_detail_by_path_with_too_many_slugs(self):
        home = Page.objects.last()
        foo = FooPage(title="foo")
        foo2 = FooPage(title="foo2")
        home.add_child(instance=foo)
        home.add_child(instance=foo2)
        foo.live = False
        foo2.live = False
        foo.save()
        foo2.save()
        foo.slug = "foo-edit"
        foo.save_revision()
        foo2.slug = "foo-edit"
        foo2.save_revision()

        url = "/api/v2/pages/detail_by_path/"
        params = {
            "html_path": "/foo-edit/",
            "draft": hash_draft_code(TEST_DRAFT_CODE, foo.pk),
        }
        res = self.client.get(url, params)

        self.assertEqual(res.status_code, 404)

    def test_sitemap_with_site(self):
        home = Page.objects.last()
        site2_hostname = "http://example.com"
        site2 = Site.objects.create(root_page=home, hostname=site2_hostname)
        params = {"site": site2.id}
        request = RequestFactory().get("sitemap.xml", params)
        res = sitemap(request)
        res.render()
        self.assertContains(res, site2_hostname)

    def test_redirect_viewset(self):
        params = {"old_path": "/lol/"}
        request = APIRequestFactory().get("", params)
        redirect_list = RedirectViewSet.as_view({"get": "list"})
        Redirect.objects.create(old_path="/lol/", redirect_link="https://example.com")
        Redirect.objects.create(old_path="/test/", redirect_link="https://no.com")
        response = redirect_list(request)
        self.assertContains(response, "example.com")
        self.assertNotContains(response, "no.com")

    def test_exclude_type(self):
        home = Page.objects.last()
        foo = FooPage(title="foo")
        home.add_child(instance=foo)
        params = {"exclude_type": "sandbox.FooPage"}
        request = APIRequestFactory().get("", params)
        request.site = Site.objects.first()
        request.wagtailapi_router = WagtailAPIRouter("wagtailapi")
        page_list = SPAExtendedPagesAPIEndpoint.as_view({"get": "listing_view"})
        res = page_list(request)
        self.assertNotContains(res, foo.title)
        self.assertEqual(res.data["meta"]["total_count"], 1)

    def test_find_view(self):
        home = Page.objects.last()
        foo = FooPage(title="foo")
        home.add_child(instance=foo)

        url = "/api/v2/pages/find/"
        params = {"html_path": "/"}
        res = self.client.get(url, params)
        self.assertEqual(res.status_code, 302)

    def test_detail_by_path(self):
        home = Page.objects.last()

        url = "/api/v2/pages/detail_by_path/"
        params = {"html_path": "/"}
        res = self.client.get(url, params)
        self.assertContains(res, home.title)



from django.conf import settings
from django.conf.urls import url
from django.http import Http404
from django.shortcuts import redirect, get_object_or_404
from django_filters import rest_framework as filters
from wagtail.api.v2.views import PagesAPIViewSet
from wagtail.api.v2.utils import BadRequestError, page_models_from_string
from wagtail.contrib.sitemaps.views import sitemap as wagtail_sitemap
from wagtail.contrib.redirects.models import Redirect
from wagtail.models import Site, Page, Revision
from rest_framework import viewsets, permissions
from rest_framework.response import Response
from .serializers import RedirectSerializer
from .utils import exclude_page_type, hash_draft_code


def filter_page_type(queryset, page_models):
    qs = queryset.none()

    for model in page_models:
        qs |= queryset.type(model)

    return qs


class SPAExtendedPagesAPIEndpoint(PagesAPIViewSet):
    """
    Wagtail preview doesn't work with a JS client

    This tweaked Pages API will serve the latest draft version of a page when
    `?draft=[draft_code]` is set.

    Added `site` (id) and `site_hostname` query parameters to filter by site.

    Added `exclude_type` to exclude wagtail page types
    """

    known_query_parameters = PagesAPIViewSet.known_query_parameters.union(
        [
            "site",
            "site_hostname",
            "exclude_type",
        ]
    )

    def check_valid_draft_code(self, page_id=None):
        """Check computed hashes for the Date + PREVIEW_DRAFT_CODE + Page ID"""
        settings_draft_code = getattr(settings, "PREVIEW_DRAFT_CODE", None)
        if settings_draft_code:
            user_draft_code = self.request.GET.get("draft")
            if page_id is None:
                page_id = self.request.parser_context["kwargs"].get("pk")
            if user_draft_code and page_id:
                settings_draft_code_hash = hash_draft_code(settings_draft_code, page_id)
                if user_draft_code == settings_draft_code_hash:
                    return True
        return False

    def detail_view(self, request, pk):
        if self.check_valid_draft_code(pk):
            # Get all, not just live
            instance = get_object_or_404(Page.objects.all(), pk=pk).specific
            instance = instance.get_latest_revision_as_object()
            serializer = self.get_serializer(instance)
            return Response(serializer.data)
        return super().detail_view(request, pk)

    def find_view(self, request):
        """
        Override to append preview GET param to redirect url
        """
        url = super().find_view(request).url
        draft_code = request.GET.get("draft")
        draft_code = getattr(settings, "PREVIEW_DRAFT_CODE", None)
        if draft_code and request.GET.get("draft") == draft_code:
            url += f"?draft={draft_code}"
        return redirect(url)

    def route(self, page, request, path_components):
        """Alternative version of Page.route that supports draft pages"""
        if path_components:
            # request is for a child of this page
            child_slug = path_components[0]
            remaining_components = path_components[1:]

            # Look for page slug first.
            # Luckily, wagtail admin will not allow even a draft to have a slug that matches a published page
            subpage = page.get_children().filter(slug=child_slug).first()
            if not subpage:
                # Look in revisions if page slug not found

                # It's possible that multiple revision slugs exist for two different pages.
                # In such a case, it picks the page with the most recent revision. The hash will then fail and 404. This is a limitation.
                # In theory we could work around this by returning all matching pages and checking the draft hash of each one
                # Merge requests welcome
                revision_qs = Revision.objects.filter(
                    object_id__in=page.get_descendants().values_list("pk"),
                    content__slug=child_slug,
                ).order_by("-created_at")
                if len(revision_qs):
                    subpage = revision_qs.first().content_object
                else:
                    raise Http404

                # Confirm exact revision slug data (contains is not exact enough)
                revision_slug = subpage.get_latest_revision().content.get("slug")
                if revision_slug != child_slug:
                    raise Http404

            return self.route(subpage, request, remaining_components)
        return page

    def detail_by_path_view(self, request):
        """
        This should work similar to find_view except that it returns the detail response instead
        of a redirect. It also supports draft codes.
        This can be useful with node, which has complications when handling redirects in a
        different manner than a web browser.
        """
        queryset = self.get_queryset()

        if request.GET.get("draft"):
            queryset = self.get_queryset(include_drafts=True)
            # We have to reimplement some of wagtail's logic to include unpublished pages
            site = self.filter_by_site()
            root_page = site.root_page.specific
            path = request.GET["html_path"]
            path_components = [component for component in path.split("/") if component]
            obj = self.route(root_page, request, path_components)
            if obj and self.check_valid_draft_code(obj.id):
                self.kwargs["pk"] = obj.pk
                return self.detail_view(request, obj.pk)

        try:
            obj = self.find_object(queryset, request)

            if obj is None:
                raise self.model.DoesNotExist

        except self.model.DoesNotExist:
            raise Http404("not found")

        self.kwargs["pk"] = obj.pk
        return self.detail_view(request, obj.pk)

    def get_queryset(self, include_drafts=False):
        """
        Override this to allow for providing drafts

        This non working shows a better intent of what we want to do

        def get_queryset(self):
            self.filter_by_site()
            queryset = super().get_queryset()
            queryset = self.exclude_page_types(queryset)
            if self.check_valid_draft_code():
                # We can't remove from a queryset
                queryset = queryset.include_drafts_instead_of_live_public()
            return queryset
        """
        # Allow pages to be filtered to a specific type
        try:
            models = page_models_from_string(
                self.request.GET.get("type", "wagtailcore.Page")
            )
        except (LookupError, ValueError):
            raise BadRequestError("type doesn't exist")

        if not models:
            models = [Page]

        if len(models) == 1:
            queryset = models[0].objects.all()
        else:
            queryset = Page.objects.all()

            # Filter pages by specified models
            queryset = filter_page_type(queryset, models)

        # Get live pages that are not in a private section
        if (
            not self.check_valid_draft_code() and include_drafts is False
        ):  # Unless draft code
            queryset = queryset.public().live()

        # Filter by site
        site = self.filter_by_site()
        if site:
            queryset = queryset.descendant_of(site.root_page, inclusive=True)
        else:
            # No sites configured
            queryset = queryset.none()

        queryset = self.exclude_page_types(queryset)
        return queryset

    def exclude_page_types(self, queryset):
        exclude_type = self.request.GET.get("exclude_type", None)
        if exclude_type is not None:
            try:
                models = page_models_from_string(exclude_type)
            except (LookupError, ValueError):
                raise BadRequestError("type doesn't exist")
            queryset = exclude_page_type(queryset, models)
        return queryset

    def filter_by_site(self):
        """
        Allow API consumer to manually specify the site
        Set query parameter `site` to the site id
        or set query parameter `site_hostname` to the hostname such as www.example.com
        """
        site = None
        site_id = self.request.GET.get("site", None)
        if site_id:
            try:
                site = Site.objects.get(id=site_id)
            except Site.DoesNotExist:
                raise BadRequestError("Site not found")
        site_hostname = self.request.GET.get("site_hostname", None)
        if site_hostname:
            try:
                site = Site.objects.get(hostname=site_hostname)
            except Site.DoesNotExist:
                raise BadRequestError("Site not found")
        if site is None:
            site = Site.find_for_request(self.request)
        self.request._wagtail_site = site
        return site

    @classmethod
    def get_urlpatterns(cls):
        urlpatterns = super().get_urlpatterns()
        urlpatterns.append(
            url(
                r"^detail_by_path/$",
                cls.as_view({"get": "detail_by_path_view"}),
                name="detail_by_path",
            )
        )
        return urlpatterns


class RedirectViewSet(viewsets.ReadOnlyModelViewSet):
    queryset = Redirect.objects.all()
    serializer_class = RedirectSerializer
    permission_classes = [permissions.AllowAny]
    filter_backends = (filters.DjangoFilterBackend,)
    filterset_fields = ("old_path", "site")
    model = Redirect

    @classmethod
    def get_urlpatterns(cls):
        """
        This returns a list of URL patterns for the endpoint
        """
        return [
            url(r"^$", cls.as_view({"get": "list"})),
        ]


def sitemap(request, sitemaps=None, **kwargs):
    """Extended wagtail sitemap view. Adds `site` query parameter to site site ID"""
    site_id = request.GET.get("site", None)
    if site_id:
        try:
            request._wagtail_site = Site.objects.get(id=site_id)
        except Site.DoesNotExist:
            pass

    return wagtail_sitemap(request, sitemaps=sitemaps, **kwargs)



# Derived from https://github.com/torchbox/wagtail-headless-preview#example

from django.contrib.contenttypes.models import ContentType

from wagtail.api.v2.views import PagesAPIViewSet

from wagtail_headless_preview.models import PagePreview
from rest_framework.response import Response


class PagePreviewAPIViewSet(PagesAPIViewSet):
    known_query_parameters = PagesAPIViewSet.known_query_parameters.union(
        ["content_type", "token"]
    )

    def listing_view(self, request):
        page = self.get_object()
        serializer = self.get_serializer(page)
        return Response(serializer.data)

    def detail_view(self, request, pk):
        page = self.get_object()
        serializer = self.get_serializer(page)
        return Response(serializer.data)

    def get_object(self):
        app_label, model = self.request.GET["content_type"].split(".")
        content_type = ContentType.objects.get(app_label=app_label, model=model)

        page_preview = PagePreview.objects.get(
            content_type=content_type, token=self.request.GET["token"]
        )
        page = page_preview.as_page()
        if not page.pk:
            # fake primary key to stop API URL routing from complaining
            page.pk = 0

        return page



from django.apps import AppConfig


class WagtailSPAIntegrationAppConfig(AppConfig):
    name = "wagtail_spa_integration"
    verbose_name = "Wagtail SPA integration"
    default_auto_field = "django.db.models.AutoField"



from django.conf import settings
from django import template
from ..utils import hash_draft_code

register = template.Library()


@register.simple_tag(takes_context=True)
def get_preview_code(context):
    page_id = context.request.resolver_match.kwargs["page_id"]
    return hash_draft_code(settings.PREVIEW_DRAFT_CODE, int(page_id))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['wagtail_spa_integration', 'wagtail_spa_integration.templatetags']

package_data = \
{'': ['*'], 'wagtail_spa_integration': ['templates/wagtailadmin/pages/*']}

install_requires = \
['Django>=3.2.0',
 'django-filter>=2.3.0,<3.0.0',
 'wagtail-headless-preview>=0.2,<0.3',
 'wagtail>=4.1.0,<5.0.0']

setup_kwargs = {
    'name': 'wagtail-spa-integration',
    'version': '3.0.0b1',
    'description': 'Tools for using Wagtail API with JavaScript single page apps',
    'long_description': 'None',
    'author': 'thelabnyc',
    'author_email': 'thelabdev@thelabnyc.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
